use anyhow::{bail, Result};
use chrono::prelude::*;
use envconfig::Envconfig;
use openlipc_dyn::Lipc;
use std::{rc::Rc, sync::Mutex, thread, time};

use kdash_client;

#[tokio::main]
async fn main() -> Result<()> {
    env_logger::init();

    let env_config = kdash_client::config::Config::init_from_env().unwrap();

    let api = kdash_client::api::Api::new(env_config.kdash_jwt, env_config.kdash_url)?;
    let config = api.fetch_config().await?;

    let lipc = Lipc::load(None)?;

    let state = kdash_client::app::State {
        api,
        app_config: kdash_client::app::AppConfig {
            net: env_config.net,
            router_ip: env_config.router_ip,
            assets_path: env_config.assets,
        },
        config: Rc::new(Mutex::new(config)),
        lipc: Rc::new(Mutex::new(lipc)),
    };

    let mut display = kdash_client::fb::FramebufferDisplay::new(
        kdash_client::fb::DEFAULT_FB,
        state.config.lock().unwrap().device.orientation,
    )?;

    kdash_client::utils::kill_kindle()?;

    let mut error_count: usize = 0;
    loop {
        display.clear();

        let mut error_suspend = false;

        match kdash_client::run_once(&state, &mut display).await {
            Ok(Some(new_config)) => {
                log::info!("Updating config");
                let mut config = state.config.lock().unwrap();
                if display.fb_orientation.orientation != new_config.device.orientation {
                    log::info!(
                        "Updating orientation: {:?} -> {:?}",
                        display.fb_orientation.orientation,
                        new_config.device.orientation
                    );
                    display.set_orientation(config.device.orientation)?;
                }
                *config = new_config;
            }
            Ok(None) => {
                log::info!("Did not fetch config");
                error_suspend = true;
            }
            Err(e) => {
                log::error!("Error during main loop: {}", e);
                error_suspend = true;
            }
        }

        let config = state.config.lock().unwrap();

        thread::sleep(config.time.delay_before_suspend.to_std()?);

        log::info!("Calculate next timer and going to sleep");

        if error_suspend {
            error_count += 1;

            log::error!("An error has occured",);

            if error_count >= 10 {
                log::info!("Reboot because of {} errors", error_count);
                kdash_client::utils::exec_command_discard("reboot", &[])?;
                bail!("Rebooting");
            }

            if config.device.use_rtc {
                kdash_client::rtc::sleep(config.time.delay_on_error.to_std()?.as_secs())?;
            } else {
                thread::sleep(config.time.delay_on_error.to_std()?);
            }
        } else {
            error_count = 0;

            log::info!("Success");

            let upcoming = config
                .time
                .update_schedule
                .upcoming(config.time.timezone)
                .next();

            let dur = match upcoming {
                Some(dt) => (dt.to_utc().naive_utc() - Utc::now().naive_utc()).to_std()?,
                None => time::Duration::from_secs(60),
            };

            if config.device.use_rtc {
                kdash_client::rtc::sleep(dur.as_secs())?;
            } else {
                thread::sleep(dur);
            }
        }
    }
}
