use anyhow::{bail, Result};
use openlipc_dyn::Lipc;
use std::fs;
use std::io::Write;
use std::process;

pub fn exec_command_discard(cmd: &str, args: &[&str]) -> Result<()> {
    log::debug!(
        "Executing command (discarding output): {:?} {:?}",
        cmd,
        args
    );

    let output = process::Command::new(cmd)
        .args(args)
        .stdout(process::Stdio::inherit())
        .stderr(process::Stdio::inherit())
        .output()?;
    if !output.status.success() {
        bail!("Command threw error: {}", output.status);
    }

    Ok(())
}

pub fn exec_command(cmd: &str, args: &[&str]) -> Result<String> {
    log::debug!("Executing command: {:?} {:?}", cmd, args);

    let output = process::Command::new(cmd).args(args).output()?;
    if !output.status.success() {
        bail!("Command threw error: {}", output.status);
    }

    let s = String::from_utf8(output.stdout)?;

    Ok(s)
}

pub fn kdb_get(name: &str) -> Result<String> {
    log::debug!("kdb get: {:?}", name);

    let mut s = exec_command("kdb", &["get", name])?;
    if s.ends_with('\n') {
        s = s.trim_end_matches('\n').to_string();
    }

    Ok(s)
}

const KILL_KINDLE_COMMANDS: &[(&str, &[&str])] = &[
    ("/etc/init.d/framework", &["stop"]),
    ("/etc/init.d/cmd", &["stop"]),
    ("/etc/init.d/phd", &["stop"]),
    ("/etc/init.d/volumd", &["stop"]),
    ("/etc/init.d/tmd", &["stop"]),
    ("/etc/init.d/webreaderd", &["stop"]),
];

const KILL_KINDLE_KILLALL: (&str, &[&str]) = ("killall", &["-q", "lipc-wait-event"]);

pub fn kill_kindle() -> Result<()> {
    log::info!("Killing Kindle");

    for (cmd, args) in KILL_KINDLE_COMMANDS {
        exec_command_discard(cmd, args)?;
    }
    let _ = exec_command_discard(KILL_KINDLE_KILLALL.0, KILL_KINDLE_KILLALL.1);

    Ok(())
}

const CPU_SCALING_GOVERNOR: &str = "/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor";
const CPU_SCALING_GOVERNOR_VALUE: &[u8] = "powersave".as_bytes();

pub fn set_cpu_powersaving() -> Result<()> {
    log::info!("Setting CPU into powersaving mode");

    let mut file = fs::OpenOptions::new()
        .write(true)
        .open(CPU_SCALING_GOVERNOR)?;
    file.write_all(CPU_SCALING_GOVERNOR_VALUE)?;

    Ok(())
}

pub fn prevent_screensaver(lipc: &Lipc) -> Result<()> {
    log::info!("Preventing screensaver");
    lipc.set_int_prop("com.lab126.powerd", "preventScreenSaver", 1)?;
    Ok(())
}
